package dev.pfaff.unfettered;

public enum CompressedOopsMode {
	ProbablyDisabled,
	UnscaledNarrow,
	ZeroBasedNarrow,
	DisjointBaseNarrow,
	HeapBasedNarrow;

	public static final CompressedOopsMode RUNNING;

	static {
		String propValue = System.getProperty("java.vm.compressedOopsMode");
		if (propValue != null) {
			RUNNING = switch (propValue) {
				case "32-bit" -> UnscaledNarrow;
				case "Zero based" -> ZeroBasedNarrow;
				case "Non-zero disjoint base" -> DisjointBaseNarrow;
				case "Non-zero based" -> HeapBasedNarrow;
				default -> throw new AssertionError(
						"Unsupported JVM: value of property java.vm.compressedOopsMode is unrecognized.");
			};
		} else {
			RUNNING = ProbablyDisabled;
		}
	}
}
