package dev.pfaff.unfettered;

public final class Log2 {
	/**
	 * Exact integer log2. See <a href="https://stackoverflow.com/a/3305710">https://stackoverflow.com/a/3305710</a>.
	 * <p>
	 * <h4>Implementations notes:</h4>
	 *
	 * Returns {@code 0} for input {@code 0}.
	 */
	public static int log2nlz(int n) {
		if (n == 0) return 0;
		if (Integer.bitCount(n) != 1) throw new IllegalArgumentException("Cannot compute exact log2 of " + n);
		return log2nlzUnchecked(n);
	}

	/**
	 * Exact long log2. See {@link #log2nlz} for details.
	 * <p>
	 * <h4>Implementations notes:</h4>
	 *
	 * Returns {@code 0} for input {@code 0}.
	 */
	public static int log2nlz(long n) {
		if (n == 0L) return 0;
		if (Long.bitCount(n) != 1) throw new IllegalArgumentException("Cannot compute exact log2 of " + n);
		return log2nlzUnchecked(n);
	}

	public static int log2nlzUnchecked(int n) {
		return 31 - Integer.numberOfLeadingZeros(n);
	}

	public static int log2nlzUnchecked(long n) {
		return 63 - Long.numberOfLeadingZeros(n);
	}
}
